import numpy as np
import numpy.random as rnd
import torch
from torchvision.transforms import Compose
from elasticdeform import deform_random_grid
from skimage.transform import resize, rotate


def split_segmentation_transforms(transforms):
    """
    Splits the transforms in a shared part, and one specifically for the inputs and outputs

    :param transforms: transform object (Compose)
    :return: subsets of transforms (Compose): shared_transforms, x_transforms and y_transforms
    """

    shared_transforms = []
    x_transforms = []
    y_transforms = []
    for transform in transforms.transforms:
        if transform.__class__ == AddNoise or transform.__class__ == Normalize or \
                transform.__class__ == ContrastAdjust:
            x_transforms.append(transform)
        elif transform.__class__ == CleanDeformedLabels:
            y_transforms.append(transform)
        else:
            shared_transforms.append(transform)
    return Compose(shared_transforms), Compose(x_transforms), Compose(y_transforms)


class ToTensor(object):
    """
    Transforms a numpy array into a tensor

    :param forward x: input array (N_1, N_2, N_3, ...)
    :return: output tensor (N_1, N_2, N_3, ...)
    """

    def __call__(self, x):
        return torch.from_numpy(x)


class ToFloat(object):
    """
    Transforms a Tensor to a FloatTensor

    :param forward x: input array (N_1, N_2, N_3, ...)
    :return: output tensor (N_1, N_2, N_3, ...)
    """

    def __call__(self, x):
        return x.float()


class ToLong(object):
    """
    Transforms a Tensor to a LongTensor

    :param forward x: input array (N_1, N_2, N_3, ...)
    :return: output tensor (N_1, N_2, N_3, ...)
    """

    def __call__(self, x):
        return x.long(),


class AddNoise(object):
    """
    Adds noise to the input

    :param initialization prob: probability of adding noise
    :param initialization sigma_min: minimum noise standard deviation
    :param initialization sigma_max: maximum noise standard deviation
    :param forward x: input array (N_1, N_2, ...)
    :return: output array (N_1, N_2, ...)
    """

    def __init__(self, prob=0.5, sigma_min=0.0, sigma_max=1.0):
        self.prob = prob
        self.sigma_min = sigma_min
        self.sigma_max = sigma_max

    def __call__(self, x):

        if rnd.rand() < self.prob:
            sigma = rnd.uniform(self.sigma_min, self.sigma_max)
            noise = rnd.randn(*x.shape) * sigma
            return x + noise
        else:
            return x


class Normalize(object):
    """
    Normalizes the input

    :param type: the desired type of normalization (z, unit or minmax)
    :param optional bits: number of bits used to represent a pixel value (only if type is unit)
    :param optional mu: normalization mean (only if type is z)
    :param optional sigma: normalization std (only if type is z)
    :param forward x: input array (N_1, N_2, ...)
    :return: output array (N_1, N_2, N_3, ...)
    """

    def __init__(self, type='unit', bits=8, mu=None, sigma=None):
        self.type = type
        self.bits = bits
        self.mu = mu
        self.sigma = sigma

    def __call__(self, x):
        if self.type == 'minmax':
            # apply minmax normalization
            m = np.min(x)
            M = np.max(x)
            eps = 1e-5
            return (x - m + eps) / (M - m + eps)
        elif self.type == 'unit':
            # apply unit normalization
            return x / (2**self.bits)
        else:
            # apply z normalization
            mu = np.mean(x) if self.mu is None else self.mu
            sigma = np.std(x) if self.sigma is None else self.sigma
            return (x - mu) / sigma


class ContrastAdjust(object):
    """
    Apply contrast adjustments to the data

    :param initialization prob: probability of adjusting contrast
    :param initialization adj: maximum adjustment (maximum intensity shift for minimum and maximum the new histogram)
    :param forward x: input array (N_1, N_2, N_3, ...)
    :return: output array (N_1, N_2, N_3, ...)
    """

    def __init__(self, prob=1, adj=0.1):
        self.prob = prob
        self.adj = adj

    def __call__(self, x):

        if rnd.rand() < self.prob:
            x_ = x

            m = np.min(x_)
            M = np.max(x_)
            r1 = rnd.rand()
            r2 = rnd.rand()
            m_ = 2 * self.adj * r1 - self.adj + m
            M_ = 2 * self.adj * r2 - self.adj + M

            if m != M:
                return ((x - m) / (M - m)) * (M_ - m_) + m_
            else:
                return x
        else:
            return x


class Scale(object):
    """
    Scales the input by a specific factor (randomly selected from a minimum-maximum range)

    :param initialization dsize: target size [Z' ,] Y' , X'
    :param initialization mode: padding type (‘constant’, ‘edge’, ‘symmetric’, ‘reflect’, ‘wrap’)
    :param forward x: input array (C, [Z  ,] Y  , X)
    :return: output tensor (C, [Z' ,] Y' , X')
    """

    def __init__(self, dsize, mode='edge'):
        self.dsize = dsize
        self.mode = mode

    def __call__(self, x):
        x_ = np.zeros_like(x)
        for c in range(x.shape[0]):
            x_[c] = resize(x[c], self.dsize, mode=self.mode)
        return x_


class Flip(object):
    """
    Perform a flip along a specific dimension

    :param initialization prob: probability of flipping
    :param initialization dim: (spatial) dimension to flip (excluding the channel dimension, default flips last dim)
    :param forward x: input array (C, N_1, N_2, ...)
    :return: output array (C, N_1, N_2, ...)
    """

    def __init__(self, prob=0.5, dim=None):
        self.prob = prob
        if dim is None:
            self.dim = -1
        else:
            self.dim = dim + 1

    def __call__(self, x):

        if rnd.rand() < self.prob:
            return np.flip(x, axis=self.dim).copy()
        else:
            return x


class Rotate90(object):
    """
    Rotate the inputs by 90 degree angles

    :param initialization prob: probability of rotating
    :param forward x: input array (C, N_1, N_2, ...)
    :return: output array (C, N_1, N_2, ...)
    """

    def __init__(self, prob=1):
        self.prob = prob

    def __call__(self, x):

        if rnd.rand() < self.prob:
            n = x.ndim
            return np.rot90(x, k=rnd.randint(0, 4), axes=[n - 2, n - 1]).copy()
        else:
            return x


class RotateRandom(object):
    """
    Rotate the inputs by a random amount of degrees within interval.

    :param initialization angle: max rotation angle (sampled symmetrically around 0)
    :param initialization resize: resize to fit
    :param initialization mode: padding type (‘constant’, ‘edge’, ‘symmetric’, ‘reflect’, ‘wrap’)
    :param forward x: input array (C, [Z ,] Y , X)
    :return: output array (C, [Z ,] Y , X)
    """

    def __init__(self, prob=1.0, angle=30, resize=False, mode='edge'):
        self.prob = prob
        self.angle = angle
        self.resize = resize
        self.mode = mode

    def __call__(self, x):

        if rnd.rand() < self.prob:
            angle = - self.angle + rnd.rand() * 2 * self.angle
            x_ = np.zeros_like(x)
            for c in range(x.shape[0]):
                x_[c] = rotate(x[c], angle, resize=self.resize, mode=self.mode)
            return x_
        else:
            return x


class RandomCrop(object):
    """
    Selects a random crop from the input

    :param initialization crop_shape: shape of the crop [Z' ,] Y' , X'
    :param forward x: input array (C, [Z ,] Y , X)
    :return: output tensor (C, [Z' ,] Y' , X')
    """

    def __init__(self, crop_shape):
        self.crop_shape = crop_shape

    def __call__(self, x):
        z_ = np.random.randint(0, x.size(1) - self.crop_shape[0] + 1)
        y_ = np.random.randint(0, x.size(2) - self.crop_shape[1] + 1)
        if len(self.crop_shape) == 2:  # 2D
            return x[:, z_:z_ + self.crop_shape[0], y_:y_ + self.crop_shape[1]]
        else:  # 3D
            x_ = np.random.randint(0, x.size(3) - self.crop_shape[2] + 1)
            return x[:, z_:z_ + self.crop_shape[0], y_:y_ + self.crop_shape[1], x_:x_ + self.crop_shape[2]]


class RandomDeformation(object):
    """
    Apply random deformation to the inputs

    :param initialization prob: probability of deforming the data
    :param initialization sigma: standard deviation of the normal deformation distribution
    :param initialization points: number of points of the deformation grid
    :param initialization mode: padding type (‘constant’, ‘edge’, ‘symmetric’, ‘reflect’, ‘wrap’)
    :param forward x: input array (C, [Z ,] Y , X)
    :return: output tensor (C, [Z ,] Y , X)
    """

    def __init__(self, prob=1, sigma=10, points=5, mode='nearest'):
        self.prob = prob
        self.sigma = sigma
        self.points = points
        self.mode = mode

    def __call__(self, x):

        if rnd.rand() < self.prob:
            if x.ndim == 3:  # 2D
                return deform_random_grid(x, sigma=self.sigma, points=self.points, axis=(1, 2), mode=self.mode)
            else:  # 3D
                return deform_random_grid(x, sigma=self.sigma, points=self.points, axis=(1, 2, 3), mode=self.mode)
        else:
            return x


class CleanDeformedLabels(object):
    """
    Clean the deformed labels by mapping the floating point values to nearest integers

    :param initialization coi: classes of interest
    :param forward x: input array (C, [Z ,] Y , X)
    :return: output tensor (C, [Z ,] Y , X)
    """

    def __init__(self, coi):
        self.coi = np.arange(len(coi))

    def __call__(self, x):

        d = np.zeros((len(self.coi), *x.shape), dtype=x.dtype)
        for i, c in enumerate(self.coi):
            d[i] = np.abs(x - c)
        d = np.argmin(d, axis=0)

        for i, c in enumerate(self.coi):
            x[d == i] = c

        return x



import numpy as np
import numpy.random as rnd
import torch
import torch.nn.functional as F
import torch.utils.data as data

from neuralnets.util.io import read_volume, print_frm


def _len_epoch(input_shape, data_shape, target=0.01, delta = 0.005, df=2, max_len=1000000, reps=20, D_max=100000):

    def _compute_target(d, D, reps, n):
        cnt = 0
        for i in range(reps):
            x = np.zeros((D))
            for k in range(n):
                j = rnd.randint(D - d)
                x[j:j+d] = 1
            cnt += np.mean(1-x)
        target = cnt / reps
        return target

    rnd.seed(0)

    d = np.prod(input_shape)
    D = np.prod(data_shape)

    r = np.ceil(np.power(D / D_max, 1 / 3))
    d = max(1, int(d / r / r / r))
    D = int(D / r / r / r)

    n = 1
    f = 1 + df
    t = 1
    while n < max_len:
        t = _compute_target(d, D, reps, n)
        if np.abs(t - target) < delta:
            return t, n
        elif t >= target + delta:
            n = min(max(n + 1, int(n * f)), max_len)
        else:
            n = max(min(n - 1, int(n / f)), 1)
            df = df / 2
            f = 1 + df

    return t, n


def load_data(x, data_type='pngseq', dtype='uint8'):
    if isinstance(x, str):  # path to a single volume
        return read_volume(x, type=data_type, dtype=dtype)
    elif isinstance(x, np.ndarray):  # data already loaded
        return x
    else:
        raise ValueError('Datasets require data in str or np.ndarray format, received %s' % type(x))


def pad2multiple(x, input_shape, value=0):

    pad_width = []
    for d in range(3):
        ts = (int(np.ceil(x.shape[d] / input_shape[d])) * input_shape[d]) - x.shape[d]
        pad_width.append((0, ts))
    pad_width = tuple(pad_width)

    x = np.pad(x, pad_width=pad_width, mode='constant', constant_values=value)

    return x


def pad_channels(x, in_channels=1, orientations=(0,)):

    pad_width = []
    for d in range(3):
        ts = 0
        if d in orientations:
            ts = in_channels - 1
        pad_width.append((0, ts))
    pad_width = tuple(pad_width)

    x = np.pad(x, pad_width=pad_width, mode='edge')

    return x


def slice_subset(x, range, orientation):
    if range is not None and orientation is not None:
        start, stop = range
        z_dim, y_dim, x_dim = x.shape
        if orientation == 'z':
            start = int(start * z_dim)
            stop = int(stop * z_dim)
            x = x[start:stop, :, :]
        elif orientation == 'y':
            start = int(start * y_dim)
            stop = int(stop * y_dim)
            x = x[:, start:stop, :]
        else:
            start = int(start * x_dim)
            stop = int(stop * x_dim)
            x = x[:, :, start:stop]
    return x


class StandardDataset(data.Dataset):
    """
    Standard dataset of N 2D images

    :param data_path: path to the dataset
    :param optional scaling: tuple used for rescaling the data, or None
    :param optional type: type of the volume file (tif2d, tif3d, tifseq, hdf5, png or pngseq)
    :param optional dtype: type of the data (typically uint8)
    :param optional norm_type: type of the normalization (unit, z or minmax)
    """

    def __init__(self, data_path, scaling=None, type='tif3d', dtype='uint8', norm_type='unit'):
        self.data_path = data_path
        self.scaling = scaling
        self.norm_type = norm_type

        # load the data
        self.data = read_volume(data_path, type=type, dtype=dtype)

        # rescale the dataset if necessary
        if scaling is not None:
            target_size = np.asarray(np.multiply(self.data.shape, scaling), dtype=int)
            self.data = \
                F.interpolate(torch.Tensor(self.data[np.newaxis, np.newaxis, ...]), size=tuple(target_size),
                              mode='area')[0, 0, ...].numpy()

    def __getitem__(self, i):
        pass

    def __len__(self):
        return self.data.shape[0]

    def _get_stats(self):
        mu = np.mean(self.data)
        std = np.std(self.data)

        return mu, std


class VolumeDataset(data.Dataset):
    """
    Dataset for volumes

    :param data: input data, possible formats
            - path to the dataset (string)
            - preloaded 3D volume (numpy array)
            - list of paths to multiple datasets (list of strings)
            - list of preloaded 3D volumes (list of numpy arrays)
    :param input_shape: 3-tuple that specifies the input shape for sampling
    :param optional scaling: 3-tuple used for rescaling the data, or a list of 3-tuples in case of multiple datasets
    :param optional len_epoch: number of iterations for one epoch
    :param optional type: type of the volume file (tif2d, tif3d, tifseq, hdf5, png or pngseq)
    :param optional in_channels: amount of subsequent slices to be sampled (only for 2D sampling)
    :param optional orientations: list of orientations for sampling
    :param optional batch_size: size of the sampling batch
    :param optional dtype: type of the data (typically uint8)
    :param optional norm_type: type of the normalization (unit, z or minmax)
    :param optional range_split: range of slices (start, stop) to select (normalized between 0 and 1), or a list of
                                 2-tuples in case of multiple datasets
    :param optional range_dir: orientation of the slicing or a list of orientations in case of multiple datasets
    :param optional resolution: list of 3-tuples specifying the pixel resolution of the data
    :param optional match_resolution_to: match the resolution of all data to a specific dataset
    :param optional sampling_type: type of sampling in case of multiple datasets
            - joint: the dataset will generate random samples in each dataset and return all of them
            - single: the dataset will generate a random sample from a randomly selected dataset and return that
    """

    def __init__(self, data, input_shape, scaling=None, len_epoch=None, type='tif3d', in_channels=1, orientations=(0,),
                 batch_size=1, dtype='uint8', norm_type='unit', range_split=None, range_dir=None, resolution=None,
                 match_resolution_to=None, sampling_type='joint'):

        if isinstance(data, str) or isinstance(data, np.ndarray):
            self.data = [load_data(data, data_type=type, dtype=dtype)]
        elif isinstance(data, list) or isinstance(data, tuple):  # list of data
            self.data = []
            for data_i in data:
                self.data.append(load_data(data_i, data_type=type, dtype=dtype))
        else:
            raise ValueError('LabeledVolumeDataset requires data in str, np.ndarray or list format')
        self.input_shape = input_shape
        self.scaling = scaling
        self.len_epoch = int(len_epoch) if len_epoch is not None else len_epoch
        self.in_channels = int(in_channels)
        if isinstance(orientations, str):
            self.orientations = [int(item) for item in orientations.split(',')]
        else:
            self.orientations = orientations
        self.orientation = 0
        self.batch_size = int(batch_size)
        self.norm_type = norm_type
        self.range_split = range_split
        self.range_dir = range_dir
        self.resolution = resolution
        self.match_resolution_to = match_resolution_to
        self.sampling_type = sampling_type
        self.warned = False

        # select a subset of slices of the data
        for i in range(len(self.data)):
            if isinstance(range_dir, list) or isinstance(range_dir, tuple):
                self.data[i] = slice_subset(self.data[i], range_split[i], range_dir[i])
            else:
                self.data[i] = slice_subset(self.data[i], range_split, range_dir)

        # compute length epoch if necessary
        if len_epoch is None or len_epoch < 0:
            print_frm('Epoch length not set... estimating full coverage value')
            if self.sampling_type == 'single':
                tt, tle = 0, 0
                for d in self.data:
                    t, le = _len_epoch(self.input_shape, d.shape)
                    tt += t
                    tle += le
                t, self.len_epoch = tt / len(self.data), tle
            else:
                i_max = int(np.argmax([d.size for d in self.data]))
                t, self.len_epoch = _len_epoch(self.input_shape, self.data[i_max].shape)
            print_frm('Epoch length automatically set to %d, this covers %.2f%% of the largest data on average'
                      % (self.len_epoch, (1-t)*100))

        # rescale the dataset if necessary
        for i in range(len(self.data)):
            if (self.resolution is not None and self.match_resolution_to is not None) or self.scaling is not None:
                if self.resolution is not None and self.match_resolution_to is not None:
                    scale_factor = np.divide(self.data[self.match_resolution_to].shape, self.data[i].shape)
                else:
                    if isinstance(self.scaling, list):
                        scale_factor = self.scaling[i]
                    else:
                        scale_factor = self.scaling
                data_it = torch.Tensor(self.data[i]).unsqueeze(0).unsqueeze(0)
                self.data[i] = F.interpolate(data_it, scale_factor=scale_factor, mode='bilinear')[0, 0, ...].numpy()

    def __getitem__(self, i):
        pass

    def __len__(self):
        return self.len_epoch

    def _select_orientation(self):
        self.orientation = np.random.choice(self.orientations)


class SlidingWindowDataset(data.Dataset):
    """
    Dataset for sliding window over volumes

    :param data: input data, possible formats
            - path to the dataset (string)
            - preloaded 3D volume (numpy array)
            - list of paths to multiple datasets (list of strings)
            - list of preloaded 3D volumes (list of numpy arrays)
    :param input_shape: 3-tuple that specifies the input shape for sampling
    :param optional scaling: 3-tuple used for rescaling the data, or a list of 3-tuples in case of multiple datasets
    :param optional type: type of the volume file (tif2d, tif3d, tifseq, hdf5, png or pngseq)
    :param optional in_channels: amount of subsequent slices to be sampled (only for 2D sampling)
    :param optional orientations: list of orientations for sampling
    :param optional batch_size: size of the sampling batch
    :param optional dtype: type of the data (typically uint8)
    :param optional norm_type: type of the normalization (unit, z or minmax)
    :param optional range_split: range of slices (start, stop) to select (normalized between 0 and 1), or a list of
                                 2-tuples in case of multiple datasets
    :param optional range_dir: orientation of the slicing or a list of orientations in case of multiple datasets
    :param optional resolution: list of 3-tuples specifying the pixel resolution of the data
    :param optional match_resolution_to: match the resolution of all data to a specific dataset
    """

    def __init__(self, data, input_shape, scaling=None, type='tif3d', in_channels=1, orientations=(0,), batch_size=1,
                 dtype='uint8', norm_type='unit', range_split=None, range_dir=None, resolution=None,
                 match_resolution_to=None):

        if isinstance(data, str) or isinstance(data, np.ndarray):
            self.data = [load_data(data, data_type=type, dtype=dtype)]
        elif isinstance(data, list) or isinstance(data, tuple):  # list of data
            self.data = []
            for data_i in data:
                self.data.append(load_data(data_i, data_type=type, dtype=dtype))
        else:
            raise ValueError('SlidingWindowDataset requires data in str, np.ndarray or list format')
        self.input_shape = input_shape
        self.scaling = scaling
        self.in_channels = int(in_channels)
        if isinstance(orientations, str):
            self.orientations = [int(item) for item in orientations.split(',')]
        else:
            self.orientations = orientations
        self.orientation = 0
        self.batch_size = int(batch_size)
        self.batch_size = batch_size
        self.norm_type = norm_type
        self.range_split = range_split
        self.range_dir = range_dir
        self.resolution = resolution
        self.match_resolution_to = match_resolution_to
        self.warned = False

        # select a subset of slices of the data
        for i in range(len(self.data)):
            if isinstance(range_dir, list) or isinstance(range_dir, tuple):
                self.data[i] = slice_subset(self.data[i], range_split[i], range_dir[i])
            else:
                self.data[i] = slice_subset(self.data[i], range_split, range_dir)

        # rescale the dataset if necessary
        for i in range(len(self.data)):
            if (self.resolution is not None and self.match_resolution_to is not None) or self.scaling is not None:
                if self.resolution is not None and self.match_resolution_to is not None:
                    scale_factor = np.divide(self.data[self.match_resolution_to].shape, self.data[i].shape)
                else:
                    if isinstance(self.scaling, list):
                        scale_factor = self.scaling[i]
                    else:
                        scale_factor = self.scaling
                data_it = torch.Tensor(self.data[i]).unsqueeze(0).unsqueeze(0)
                self.data[i] = F.interpolate(data_it, scale_factor=scale_factor, mode='bilinear')[0, 0, ...].numpy()

        # pad data so that the dimensions are a multiple of the inputs shapes
        self.data = [pad2multiple(d, input_shape, value=0) for d in self.data]

        # compute size of the data
        self.n_samples_dim = np.zeros((len(self.data), 3), dtype=int)
        for i, data_i in enumerate(self.data):
            for d in range(3):
                self.n_samples_dim[i, d] = data_i.shape[d] // self.input_shape[d]
        self.n_samples = self.n_samples_dim.prod(axis=1).sum()

        # pad data so that the dimensions are a multiple of the inputs shapes
        self.data = [pad_channels(d, in_channels=in_channels, orientations=self.orientations) for d in self.data]

    def __getitem__(self, i):
        pass

    def __len__(self):
        return self.n_samples



import torch
import torch.nn as nn


class Linear(nn.Module):
    """
    Linear layer

    :param in_channels: number of input channels
    :param out_channels: number of output channels
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional norm: specify normalization ("batch", "instance" or None)
    """

    def __init__(self, in_channels, out_channels, dropout=0.0, activation=None, norm=None):
        super().__init__()

        # initialize linear layer
        self.unit = nn.Sequential(nn.Linear(in_channels, out_channels))

        # apply normalization
        if norm == 'batch':
            self.unit.add_module('norm', nn.BatchNorm1d(int(out_channels)))
        elif norm == 'instance':
            self.unit.add_module('norm', nn.InstanceNorm1d(int(out_channels)))

        # apply dropout
        if dropout > 0.0:
            self.unit.add_module('dropout', nn.Dropout(p=dropout))

        # apply activation
        if activation == 'relu':
            self.unit.add_module('activation', nn.ReLU())
        elif activation == 'sigmoid':
            self.unit.add_module('activation', nn.Sigmoid())

    def forward(self, inputs):

        return self.unit(inputs)


class Conv2D(nn.Module):
    """
    2D convolutional layer

    :param in_channels: number of input channels
    :param out_channels: number of output channels
    :param optional kernel_size: kernel size of the convolutions
    :param optional stride: stride of the convolutions
    :param optional padding: padding used for border cases ("SAME", "VALID" or None)
    :param optional bias: use bias term or not
    :param optional dilation: dilation of the convolution
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional norm: specify normalization ("batch", "instance" or None)
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, padding='SAME', bias=True, dilation=1,
                 dropout=0.0, activation=None, norm=None):
        super().__init__()

        if padding == 'SAME':
            p = kernel_size // 2
        else:  # VALID (no) padding
            p = 0

        # initialize convolutional block
        self.unit = nn.Sequential(nn.Conv2d(int(in_channels), int(out_channels), kernel_size=kernel_size,
                                            padding=p, stride=stride, bias=bias, dilation=dilation))

        # apply normalization
        if norm == 'batch':
            self.unit.add_module('norm', nn.BatchNorm2d(int(out_channels)))
        elif norm == 'instance':
            self.unit.add_module('norm', nn.InstanceNorm2d(int(out_channels)))

        # apply dropout
        if dropout > 0.0:
            self.unit.add_module('dropout', nn.Dropout2d(p=dropout))

        # apply activation
        if activation == 'relu':
            self.unit.add_module('activation', nn.ReLU())
        elif activation == 'sigmoid':
            self.unit.add_module('activation', nn.Sigmoid())

    def forward(self, inputs):

        return self.unit(inputs)


class DenseLayer2D(nn.Module):
    """
    2D Dense convolutional layer

    :param in_channels: number of input channels
    :param k: how many filters to add each layer
    :param bn_size: multiplicative factor for number of bottle neck layers
    :param optional kernel_size: kernel size of the convolutions
    :param optional padding: padding used for border cases ("SAME", "VALID" or None)
    :param optional bias: use bias term or not
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional norm: specify normalization ("batch", "instance" or None)
    """

    def __init__(self, in_channels, k, bn_size, kernel_size=3, padding='SAME', bias=False, dropout=0.0, activation=None,
                 norm=None):
        super().__init__()

        self.conv1 = Conv2D(in_channels, bn_size * k, kernel_size=1, bias=bias, activation=activation, norm=norm)
        self.conv2 = Conv2D(bn_size * k, k, kernel_size=kernel_size, bias=bias, padding=padding, activation=activation,
                            norm=norm, dropout=dropout)

    def forward(self, x):
        if isinstance(x, torch.Tensor):
            x = [x]
        x = torch.cat(x, 1)
        x = self.conv2(self.conv1(x))
        return x


class Conv3D(nn.Module):
    """
    3D convolutional layer

    :param in_channels: number of input channels
    :param out_channels: number of output channels
    :param optional kernel_size: kernel size of the convolutions
    :param optional stride: stride of the convolutions
    :param optional padding: padding used for border cases ("SAME", "VALID" or None)
    :param optional bias: use bias term or not
    :param optional dilation: dilation of the convolution
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional norm: specify normalization ("batch", "instance" or None)
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, padding='SAME', bias=True, dilation=1,
                 dropout=0.0, activation=None, norm=None):
        super().__init__()

        if padding == 'SAME':
            p = kernel_size // 2
        else:  # VALID (no) padding
            p = 0

        # initialize convolutional block
        self.unit = nn.Sequential(nn.Conv3d(int(in_channels), int(out_channels), kernel_size=kernel_size,
                                            padding=p, stride=stride, bias=bias, dilation=dilation))

        # apply normalization
        if norm == 'batch':
            self.unit.add_module('norm', nn.BatchNorm3d(int(out_channels)))
        elif norm == 'instance':
            self.unit.add_module('norm', nn.InstanceNorm3d(int(out_channels)))

        # apply dropout
        if dropout > 0.0:
            self.unit.add_module('dropout', nn.Dropout3d(p=dropout))

        # apply activation
        if activation == 'relu':
            self.unit.add_module('activation', nn.ReLU())
        elif activation == 'sigmoid':
            self.unit.add_module('activation', nn.Sigmoid())

    def forward(self, inputs):

        return self.unit(inputs)


class DenseLayer3D(nn.Module):
    """
    3D Dense convolutional layer

    :param in_channels: number of input channels
    :param k: how many filters to add each layer
    :param bn_size: multiplicative factor for number of bottle neck layers
    :param optional kernel_size: kernel size of the convolutions
    :param optional padding: padding used for border cases ("SAME", "VALID" or None)
    :param optional bias: use bias term or not
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional norm: specify normalization ("batch", "instance" or None)
    """

    def __init__(self, in_channels, k, bn_size, kernel_size=3, padding='SAME', bias=False, dropout=0.0, activation=None,
                 norm=None):
        super().__init__()

        self.conv1 = Conv3D(in_channels, bn_size * k, kernel_size=1, bias=bias, activation=activation, norm=norm)
        self.conv2 = Conv3D(bn_size * k, k, kernel_size=kernel_size, bias=bias, padding=padding, activation=activation,
                            norm=norm, dropout=dropout)

    def forward(self, x):
        if isinstance(x, torch.Tensor):
            x = [x]
        x = torch.cat(x, 1)
        x = self.conv2(self.conv1(x))
        return x


class UNetConvBlock2D(nn.Module):
    """
    2D convolutional block of the classical U-Net (i.e. two sequential convolutional layers)

    :param in_channels: number of input channels
    :param out_channels: number of output channels
    :param optional kernel_size: kernel size of the convolutions
    :param optional padding: padding used for border cases ("SAME", "VALID" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional norm: specify normalization ("batch", "instance" or None)
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, padding='SAME', norm='batch', dropout=0.0,
                 activation='relu'):
        super().__init__()

        self.conv1 = Conv2D(in_channels, out_channels, kernel_size=kernel_size, padding=padding,
                            dropout=dropout, activation=activation, norm=norm)
        self.conv2 = Conv2D(out_channels, out_channels, kernel_size=kernel_size, padding=padding,
                            dropout=dropout, activation=activation, norm=norm)

    def forward(self, inputs):
        return self.conv2(self.conv1(inputs))


class DenseBlock2D(nn.Module):
    """
    2D dense convolutional block (derived from DenseNet)

    :param in_channels: number of input channels
    :param out_channels: number of output channels
    :param num_layers: number of dense layers
    :param k: how many filters to add each layer
    :param bn_size: multiplicative factor for number of bottle neck layers
    :param optional kernel_size: kernel size of the convolutions
    :param optional padding: padding used for border cases ("SAME", "VALID" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional norm: specify normalization ("batch", "instance" or None)
    """

    def __init__(self, in_channels, out_channels, num_layers, k, bn_size, kernel_size=3, padding='SAME', bias=False,
                 dropout=0.0, activation=None, norm=None):
        super().__init__()

        self.num_layers = num_layers

        # dense layers
        for i in range(num_layers):
            layer = DenseLayer2D(in_channels + i * k, k, bn_size, kernel_size=kernel_size, padding=padding, bias=bias,
                                 dropout=dropout, activation=activation, norm=norm)
            self.add_module('dense%d' % (i + 1), layer)

        # transition layer
        self.transition = Conv2D(in_channels + num_layers * k, out_channels, kernel_size=kernel_size, padding=padding,
                                 dropout=dropout, activation=activation, norm=norm)

    def forward(self, x):

        # dense layers
        features = [x]
        for i in range(self.num_layers):
            dense_layer = getattr(self, 'dense%d' % (i + 1))
            new_features = dense_layer(features)
            features.append(new_features)
        x = torch.cat(features, 1)

        return self.transition(x)


class UNetConvBlock3D(nn.Module):
    """
    3D convolutional block of the classical U-Net (i.e. two sequential convolutional layers)

    :param in_channels: number of input channels
    :param out_channels: number of output channels
    :param optional kernel_size: kernel size of the convolutions
    :param optional padding: padding used for border cases ("SAME", "VALID" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional norm: specify normalization ("batch", "instance" or None)
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, padding='SAME', norm='batch', dropout=0.0,
                 activation='relu'):
        super().__init__()

        self.conv1 = Conv3D(in_channels, out_channels, kernel_size=kernel_size, padding=padding,
                            dropout=dropout, activation=activation, norm=norm)
        self.conv2 = Conv3D(out_channels, out_channels, kernel_size=kernel_size, padding=padding,
                            dropout=dropout, activation=activation, norm=norm)

    def forward(self, inputs):
        return self.conv2(self.conv1(inputs))


class DenseBlock3D(nn.Module):
    """
    3D dense convolutional block (derived from DenseNet)

    :param in_channels: number of input channels
    :param out_channels: number of output channels
    :param num_layers: number of dense layers
    :param k: how many filters to add each layer
    :param bn_size: multiplicative factor for number of bottle neck layers
    :param optional kernel_size: kernel size of the convolutions
    :param optional padding: padding used for border cases ("SAME", "VALID" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional norm: specify normalization ("batch", "instance" or None)
    """

    def __init__(self, in_channels, out_channels, num_layers, k, bn_size, kernel_size=3, padding='SAME', bias=False,
                 dropout=0.0, activation=None, norm=None):
        super().__init__()

        self.num_layers = num_layers

        # dense layers
        for i in range(num_layers):
            layer = DenseLayer3D(in_channels + i * k, k, bn_size, kernel_size=kernel_size, padding=padding, bias=bias,
                                 dropout=dropout, activation=activation, norm=norm)
            self.add_module('dense%d' % (i + 1), layer)

        # transition layer
        self.transition = Conv3D(in_channels + num_layers * k, out_channels, kernel_size=kernel_size, padding=padding,
                                 dropout=dropout, activation=activation, norm=norm)

    def forward(self, x):

        # dense layers
        features = [x]
        for i in range(self.num_layers):
            dense_layer = getattr(self, 'dense%d' % (i + 1))
            new_features = dense_layer(features)
            features.append(new_features)
        x = torch.cat(features, 1)

        return self.transition(x)


class UNetUpSamplingBlock2D(nn.Module):
    """
    2D upsampling block of the classical U-Net

    :param in_channels: number of input channels
    :param out_channels: number of output channels
    :param optional deconv: us deconvolution or upsampling layers
    :param optional bias: use bias term or not
    """

    def __init__(self, in_channels, out_channels, deconv=False, bias=True):
        super().__init__()

        if deconv:  # use transposed convolution
            self.up = nn.ConvTranspose2d(in_channels, out_channels, kernel_size=2, stride=2, bias=bias)
        else:  # use bilinear upsampling
            self.up = nn.Upsample(scale_factor=2, mode='bilinear')

    def forward(self, *arg):
        if len(arg) == 2:
            return self.forward_concat(arg[0], arg[1])
        else:
            return self.forward_standard(arg[0])

    def forward_concat(self, inputs1, inputs2):

        return torch.cat([inputs1, self.up(inputs2)], 1)

    def forward_standard(self, inputs):

        return self.up(inputs)


class UNetUpSamplingBlock3D(nn.Module):
    """
    3D upsampling block of the classical U-Net

    :param in_channels: number of input channels
    :param out_channels: number of output channels
    :param optional deconv: us deconvolution or upsampling layers
    :param optional bias: use bias term or not
    """

    def __init__(self, in_channels, out_channels, deconv=False, bias=True):
        super().__init__()

        if deconv:  # use transposed convolution
            self.up = nn.ConvTranspose3d(in_channels, out_channels, kernel_size=2, stride=2, bias=bias)
        else:  # use bilinear upsampling
            self.up = nn.Upsample(scale_factor=2, mode='bilinear')

    def forward(self, *arg):
        if len(arg) == 2:
            return self.forward_concat(arg[0], arg[1])
        else:
            return self.forward_standard(arg[0])

    def forward_concat(self, inputs1, inputs2):

        return torch.cat([inputs1, self.up(inputs2)], 1)

    def forward_standard(self, inputs):

        return self.up(inputs)
